import  java.util.Properties;

import  java.util.concurrent.atomic.AtomicBoolean;

import  javax.jms.Connection;
import  javax.jms.ConnectionFactory;
import  javax.jms.Destination;
import  javax.jms.JMSException;
import  javax.jms.MapMessage;
import  javax.jms.MessageConsumer;
import  javax.jms.Session;

import  javax.naming.Context;
import  javax.naming.InitialContext;
import  javax.naming.NamingException;

import  org.apache.log4j.BasicConfigurator;
import  org.apache.log4j.ConsoleAppender;
import  org.apache.log4j.Level;
import  org.apache.log4j.Logger;
import  org.apache.log4j.PatternLayout;

import  sun.misc.Signal;
import  sun.misc.SignalHandler;

/**
 * JMS load test class
 * @author rschuetz
 */

public class LoadTestConsumer
{
 // === Static ================================================================

 private static final Logger         LOGGER          = Logger.getLogger(LoadTestSubscriber.class);

 private static final String         ID_SEQUENCEID   = "sequenceid";
 
 // === Attributes ============================================================

 private final String                JNDIFACTORYINITIAL,
                                     JNDIPROVIDERURL,
                                     JNDICONNECTIONFACTORYALIAS,
                                     JNDITOPICALIAS;

 private final long                  STARTTIME,
                                     TOTALRUNTIME;

 private final int                   COMMITINTERVAL;

 private final AtomicBoolean         STOP                 = new AtomicBoolean(false);
 
 private       Connection            connection           = null;
 private       Session               session              = null;
 private       MessageConsumer       messageConsumer      = null;

 private       long                  totalNbrMessagesRecv = 0L;
 
 // === Constructor ===========================================================

 public LoadTestConsumer(String[] args)
 {
  int argIdx;

  // Initialize logger

  BasicConfigurator.resetConfiguration();

  Logger.getRootLogger().addAppender(new ConsoleAppender(new PatternLayout("%-5p [%d{ISO8601}] %m%n")));
  Logger.getRootLogger().setLevel(Level.INFO);

  // Get parameters

  if(args.length == 6)
  {
   argIdx                     = 0;

   JNDIFACTORYINITIAL         = args[argIdx ++];
   JNDIPROVIDERURL            = args[argIdx ++];
   JNDICONNECTIONFACTORYALIAS = args[argIdx ++];
   JNDITOPICALIAS             = args[argIdx ++];

   TOTALRUNTIME               = Long   .parseLong(args[argIdx ++]);
   COMMITINTERVAL             = Integer.parseInt (args[argIdx ++]);
  }
  else
  {
   throw new IllegalArgumentException("Usage: " + LoadTestConsumer.class.getSimpleName() + 
                                      " <jndi factory initial> <jndi provider URL> <jndi connection factory alias> <jndi topic alias> " +
                                      " <total runtime in ms> " +
                                      " <commit interval in # of messages processed>");
  }

  // Initialize
  
  STARTTIME = System.currentTimeMillis();
 }

 // === Public ================================================================

 public static void main(String[] args)
 throws Exception
 {
  (new LoadTestConsumer(args)).runLoadTest();
 }
 
 public void runLoadTest()
 {
  int mainloop;

  // Init signal handler
  
  Signal.handle(new Signal("INT"), new SigIntSignalHandler());
  
  // Receive messages
  
  try
  {
   // Initial connect
 
   connect();
 
   // Publish until specified runtime elapsed
   
   for(mainloop = 0; System.currentTimeMillis() < STARTTIME + TOTALRUNTIME && !STOP.get(); mainloop ++)
   {
    receiveMessages(COMMITINTERVAL);
   }
   
   // Disconnect
   
   disconnect();
  }
  catch(Exception e)
  {
   LOGGER.error("Exception caught", e);
  }
  finally
  {
   try
   {
    disconnect(); 
   }
   catch(Exception e)
   {
    LOGGER.error("Exception caught while disconnecting", e);
   }
  }
 }
 
 private void receiveMessages(final long nbrMessages)
 throws InterruptedException, 
        JMSException
 {
  long                nbrMessagesReceived, recvMessageSequenceID, time;
  MapMessage          mapMessage;

  // Log

  LOGGER.info("Receiving " + nbrMessages + " messages");

  // Process messages

  nbrMessagesReceived   = 0L;
  time                  = System.currentTimeMillis();

  while(nbrMessagesReceived < nbrMessages && !STOP.get())
  {
   // Receive

   do
   {
    mapMessage = ((MapMessage)(messageConsumer.receive(1000)));
   } while(mapMessage == null && !STOP.get());
    
   // Continue?
   
   if(mapMessage != null)
   {
    recvMessageSequenceID = mapMessage.getLongProperty(ID_SEQUENCEID);
    
    // Update counters

    nbrMessagesReceived  ++;
    totalNbrMessagesRecv ++;
    
    if(totalNbrMessagesRecv % COMMITINTERVAL == 0 || nbrMessagesReceived == nbrMessages)
    {
     if(COMMITINTERVAL >= 1000) LOGGER.info("Committing consumer session; last sequenceID = " + recvMessageSequenceID);
     session.commit();
    }       
   }
  }

  time = Math.max(System.currentTimeMillis() - time, 1);
  
  // Log

  LOGGER.info("Done. " + nbrMessagesReceived + " messages received, speed: " + ((1000L * nbrMessagesReceived) / time) + " msgs/s");
  LOGGER.info("Totals: # msgs recv: " + totalNbrMessagesRecv);
 }

 // === Private ===============================================================

 private void connect()
 throws JMSException,
        NamingException
 {
  Properties        jndiProperties;
  InitialContext    jndiContext;
  ConnectionFactory connectionFactory;
  Destination       destination;

  // Log

  LOGGER.info("Connecting.");

  // Initialize JNDI

  jndiProperties = new Properties();
  jndiProperties.setProperty(Context.INITIAL_CONTEXT_FACTORY, JNDIFACTORYINITIAL);
  jndiProperties.setProperty(Context.PROVIDER_URL,            JNDIPROVIDERURL);

  jndiContext = new InitialContext(jndiProperties);

  // Lookup connection factory and queue

  connectionFactory = ((ConnectionFactory)(jndiContext.lookup(JNDICONNECTIONFACTORYALIAS)));
  destination       = ((Destination      )(jndiContext.lookup(JNDITOPICALIAS)));

  // Create producer connection, session and message producer

  connection = connectionFactory.createConnection();
  session    = connection.createSession(true, 0);

  // Create topic subscriber
  
  messageConsumer = session.createConsumer(destination);

  // Start connection
  
  connection.start();

  // Log

  LOGGER.info("Connected.");
 }

 private void disconnect()
 throws JMSException
 {
  // Log

  LOGGER.info("Disconnecting.");

  // Close assets

  if(messageConsumer != null) { messageConsumer.close(); messageConsumer = null; }
  if(session         != null) { session        .close(); session         = null; }
  if(connection      != null) { connection     .close(); connection      = null; }

  // Log

  LOGGER.info("Disconnected.");
 }
 
 public class SigIntSignalHandler implements SignalHandler 
 {
  public void handle(Signal signal) 
  {
   LOGGER.warn("CTRL-C received.");
   STOP.set(true);
  }
 }
}
